""" Proxies do the actual API requests. """
from abc import ABC, abstractmethod

import requests
import requests_cache
from munch import munchify

from .enloneexception import EnlOneException

class Proxy(ABC):
    """
    Proxy interface.
    """
    @abstractmethod
    def get(self, endpoint, params): pass
    @abstractmethod
    def post(self, endpoint, json): pass

class KeyProxy(Proxy):
    """
    Proxy implementation for authentication using API keys.
    """
    def __init__(self, base_url, apikey, cache=0):
        self._apikey = apikey
        self._base_url = base_url
        if cache > 0:
            requests_cache.install_cache('apikey_cache', backend='sqlite', expire_after=cache)
    def get(self, endpoint, params={}):
        """
        Do a get request adding the apikey as a parameter.
        """
        url = self._base_url + endpoint
        params["apikey"] = self._apikey
        try:
            response = requests.get(url, params=params)
        except requests.exceptions.RequestException:
            raise EnlOneException("Error contacting enl.one servers.")
        if response and response.json()["status"] == "ok":
            return munchify(response.json()["data"])
        else:
            raise EnlOneException("enl.one API call error.")
    def post(self, endpoint, json):
        """
        Do a post request adding the apikey as a parameter.
        """
        url = self._base_url + endpoint
        try:
            response = requests.post(url, params={"apikey" : self._apikey}, json=json)
        except requests.exceptions.RequestException:
            raise EnlOneException("Error contacting enl.one servers.")
        if response and response.json()["status"] == "ok":
            return munchify(response.json()["data"])
        else:
            raise EnlOneException("enl.one API call error.")

class TokenProxy(Proxy):
    """
    Proxy implementation for authentication using OAuth tokens.
    """
    def __init__(self, base_url, token, cache=0):
        self._token = token
        self._base_url = base_url + "/oauth"
        if cache > 0:
            requests_cache.install_cache('token_cache', backend='sqlite', expire_after=cache)
    def get(self, endpoint, params={}):
        """
        Do a get request adding the Authorization header.
        """
        url = self._base_url + endpoint
        headers = {'Authorization':'Bearer ' + self._token}
        try:
            response = requests.get(url, headers=headers, params=params)
        except requests.exceptions.RequestException:
            raise EnlOneException("Error contacting enl.one servers.")
        if response and response.json()["status"] == "ok":
            return munchify(response.json()["data"])
        else:
            raise EnlOneException("enl.one API call error.")
    def post(self, endpoint, json):
        """
        Do a get request adding the Authorization header.
        """
        url = self._base_url + endpoint
        headers = {'Authorization':'Bearer ' + self._token}
        try:
            response = requests.post(url, headers=headers, json=json)
        except requests.exceptions.RequestException:
            raise EnlOneException("Error contacting enl.one servers.")
        if response and response.json()["status"] == "ok":
            return munchify(response.json()["data"])
        else:
            raise EnlOneException("enl.one API call error.")

class OpenProxy:
    """
    Proxy implementation for the open API.
    """
    def get(self, endpoint):
        """
        Do a get request with no auth.
        """
        url = "https://v.enl.one/OpenApi" + endpoint
        try:
            response = requests.get(url)
        except requests.exceptions.RequestException:
            raise EnlOneException("Error contacting enl.one servers.")
        if response and response.json()["status"] == "ok":
            return munchify(response.json()["data"])
        else:
            raise EnlOneException("enl.one API call error.")



"""
Implementation of oauth specific endpoints.
"""
class VOAuth:
    """
    Implementation of oauth specific endpoints.
    """
    def __init__(self, proxy):
        self._proxy = proxy
    def profile(self):
        return self._proxy.get("/api/v1/profile")
    def googledata(self):
        return self._proxy.get("/api/v1/googledata")
    def email(self):
        return self._proxy.get("/api/v1/email")
    def telegram(self):
        return self._proxy.get("/api/v1/telegram")



"""
Implements V methods for agent information. Constructor must be provided with eihter
    apikey: an apikey from https://v.enl.one/apikey or
    token: an OAuth2 token with enough scopes for your use
as a keyword argument.

More detailed documentation on each method on
    https://v.enl.one/apikey and
    https://v.enl.one/oauth/clients
"""

from ._proxy import TokenProxy, KeyProxy, OpenProxy
from ._oauth import VOAuth
from .enloneexception import EnlOneException

def banned(gid):
    """
    Returns True iff the given google id correspond to an agent marked on V as banned.
    Open without Authentication.
    """
    return OpenProxy().get("/banned/" + gid)

class V:
    """
    Implements V methods for agent information. Constructor must be provided with eihter
        apikey: an apikey from https://v.enl.one/apikey or
        token: an OAuth2 token with enough scopes for your use
    as a keyword argument.

    More detailed documentation on each method on
        https://v.enl.one/apikey and
        https://v.enl.one/oauth/clients
    """
    _base_url = "https://v.enl.one"
    def __init__(self, cache=0, **kwargs):
        if "token" in kwargs:
            self._proxy = TokenProxy(self._base_url, kwargs["token"], cache=cache)
            self._oauth = VOAuth(self._proxy)
        elif "apikey" in kwargs:
            self._proxy = KeyProxy(self._base_url, kwargs["apikey"], cache=cache)
        else:
            raise EnlOneException("You need to either provide token or apikey as keyword argument.")

    # v1 general endpoints
    def trust(self, enlid):
        """
        V-Points and V-Level could be queried by enlid.
        If you are VL2 or higher, you can also query V-Points and -Level using the Google+ ID instead of ENLID.
        Both API key and OAuth should work with this method.
        """
        return self._proxy.get("/api/v1/agent/" + enlid + "/trust")
    def search(self, **kwargs):
        """
        Agents could be found by using this method. Several keywordarguments could be set to restrict the results.
        To receive some results either query or lat/lon must be set.
        Both API key and OAuth should work with this method.
        """
        return self._proxy.get("/api/v1/search", params=kwargs)
    def distance(self, enlid1, enlid2):
        """
        Like in a profile, the connections between two agents can be queried using this method.
        If you are VL2 or higher, you can also use the Google+ ID instead of ENLID.
        Both API key and OAuth should work with this method.
        """
        return self._proxy.get("/api/v1/agent/" + enlid1 + "/" + enlid2).hops
    def bulk_info(self, ids, telegramid=False, gid=False, array=False):
        """
        To reduce the amount of requests, multiple agent could be queried using this method.
        The agents could be passed as an array of enlid.
        Both API key and OAuth should work with this method.
        """
        url = "/api/v1/bulk/agent/info"
        if telegramid:
            url += "/telegramid"
        if gid:
            url += "/gid"
        if array:
            url += "/array"
        return self._proxy.post(url, ids)
    def location(self, enlid):
        """
        To retrive the location of an agent.
        Both API key and OAuth should work with this method.
        """
        return self._proxy.get("/api/v1/agent/" + enlid + "/location")
    def whoami(self):
        """
        To retrieve the data of the owner of this apikey.
        This method is API key specific.
        """
        return self._proxy.get("/api/v1/whoami")
    #TODO: profile pictures

    # v2 endpoints
    def list_teams(self):
        """
        You can list all teams of the user by using this method.
        Extension in V2: This can handle teams where members can be assigned multiple roles.
        Both API key and OAuth should work with this method.
        """
        return self._proxy.get("/api/v2/teams", )
    def team_details(self, teamid):
        """
        To retrieve a list of all members of a specific team.
        Extension in V2: This can handle teams where members can be assigned multiple roles.
        Note: You can only query your own teams! Since this API exposes much more data then the other APIs.
        Both API key and OAuth should work with this method.
        """
        return self._proxy.get("/api/v2/teams/" + str(teamid))

    # OAuth specifics
    def profile(self):
        """
        https://v.enl.one/oauth/clients
        This method is OAuth specific.
        """
        return self._oauth.profile()
    def googledata(self):
        """
        https://v.enl.one/oauth/clients
        This method is OAuth specific.
        """
        return self._oauth.googledata()
    def email(self):
        """
        This method is OAuth specific.
        """
        return self._oauth.email()
    def telegram(self):
        """
        https://v.enl.one/oauth/clients
        This method is OAuth specific.
        """
        return self._oauth.telegram()

    # Short-handers
    def search_one(self, **kwargs):
        """
        Short-hand to search for the first result.
        Both API key and OAuth should work with this method.
        """
        return self.search(**kwargs)[0]
    def is_ok(self, agent):
        """
        Given a search result, return true iff the agent is:
        verified, active, not quarantined, not blacklisted and not banned by nia.
        Both API key and OAuth should work with this method.
        """
        return agent.verified \
               and agent.active \
               and not agent.quarantine \
               and not agent.blacklisted \
               and not agent.banned_by_nia



from proxy import Proxy, TokenProxy, KeyProxy
from enloneexception import EnlOneException

class Tasks:
    _proxy : Proxy
    _base_url = "v.enl.one/"
    def __init__(self, **kwargs):
        if "token" in kwargs:
            _proxy = TokenProxy(_base_url, kwargs["token"])
        elif "apikey" in kwargs:
            _proxy = KeyProxy(_base_url, kwargs["apikey"])
        else:
            raise EnlOneException("You need to either provide token or apikey as keyword argument.")

    # OPs
    def get_op(self):
        raise EnlOneException("Not implemented yet.")
    def get_ops(self):
        raise EnlOneException("Not implemented yet.")
    def post_op(self):
        raise EnlOneException("Not implemented yet.")
    def put_op(self):
        raise EnlOneException("Not implemented yet.")
    def delete_op(self):
        raise EnlOneException("Not implemented yet.")
    def get_grant(self):
        raise EnlOneException("Not implemented yet.")
    def get_grants(self):
        raise EnlOneException("Not implemented yet.")
    def get_permissions(self):
        raise EnlOneException("Not implemented yet.")
    def get_messages(self):
        raise EnlOneException("Not implemented yet.")
    def get_message(self):
        raise EnlOneException("Not implemented yet.")
    def delete_message(self):
        raise EnlOneException("Not implemented yet.")
    def search_ops(self):
        raise EnlOneException("Not implemented yet.")
    def get_users(self):
        raise EnlOneException("Not implemented yet.")
    def sync_rocks_comm(self):
        raise EnlOneException("Not implemented yet.")

    # Tasks
    def get_tasks(self):
        raise EnlOneException("Not implemented yet.")
    def get_op_tasks(self):
        raise EnlOneException("Not implemented yet.")
    def search_tasks(self):
        raise EnlOneException("Not implemented by backend.")
    def get_task():
        raise EnlOneException("Not implemented yet.")
    def post_task():
        raise EnlOneException("Not implemented yet.")



"""
Implements enl.one APIs:
- v.enl.one
- tasks.enl.one (Soon)
- status.enl.one (Not very soon, but soon)
For more info: https://wiki.enl.one/doku.php?id=start

All methods returns Munches. That only means you access them both as:
- `some_result.some_property`
- `some_result['some_property']`
"""
from .v import V, banned
from .enloneexception import EnlOneException
__all__ = ["V", "banned", "EnlOneException"]
NAME = "pyenlone"



"""
Exception signaling that something went wrong while calling the API.
Future version of this package will add more specific exceptions,
but all will inherit from this one, so it's safe to test against it.
"""
class EnlOneException(Exception):
    """
    Exception signaling that something went wrong while calling the API.
    Future version of this package will add more specific exceptions,
    but all will inherit from this one, so it's safe to test against it.
    """
    pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyenlone",
    version="0.1.1",
    author="QuanticPotato",
    description="A python implementation of enl.one API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/potato-tools/pyenlone",
    packages=setuptools.find_packages(),
    install_requires=[
          'requests',
          'requests_cache',
          'munch',
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
